/*
 * Header    :
 * File      : RescueAgentBase.cxx
 * Auther    : Kosuke Shinoda
 * Since     :
 * Time-stamp: <2001-12-16 15:58:36 kshinoda>
 * Comment   :
 * End       :
 */
/*
 * Copyright (C) 2001 SHINODA, Kosuke. Jaist,Japan & CARC, AIST, JAPAN
 * Copyright (C) 2001 NODA, Itsuki, CARC, AIST, JAPAN
 */

#include <fstream>
#include "RescueAgentBase.hxx"

namespace RescueCivilian {
  using namespace Itk;
  using namespace Rescue;

  //--------------------------------------------------
  //
  Sexp RescueAgentBase::bodyAction_(Sexp::T_Atom, SubString("_body_action_"));
  Sexp* RescueAgentBase::bodyAction = &bodyAction_;
  Sexp RescueAgentBase::sayAction_(Sexp::T_Atom, SubString("_say_action_"));
  Sexp* RescueAgentBase::sayAction = &sayAction_;
  Sexp RescueAgentBase::chengeSituation_(Sexp::T_Atom, SubString("_situation_chenge"));
  Sexp* RescueAgentBase::chengeSituation = &chengeSituation_;


  //--------------------------------------------------
  //
  void RescueAgentBase::init(){
  };
    
  
  //--------------------------------------------------
  // PositTable Operator
  Bool RescueAgentBase::find(Sexp* posit_name) {
    if(posit_name->isSymbol()) {
      PositTable::const_iterator it = positTable()->find(posit_name->symVal());
      if(it != positTable()->end()){
	return True;
      }
    }
    return False;
  };

  //--------------------------------------------------
  // PositTable
  void RescueAgentBase::setPositTable(const char* filename, Sexp::Heap* heap){
    std::ifstream from(filename);
    if(!from){
      ITK_ERR("Illigal filename : " << filename);
      throw 0;
      return;
    }
    return setPositTable(from, heap);
  };
  void RescueAgentBase::setPositTable(istream* from, Sexp::Heap* heap){
    Sexp* posits = heap->scan(from);
    posits->fixTag();
    nop();
    return setPositTable(posits, heap);
  };
  void RescueAgentBase::setPositTable(Sexp* posits, Sexp::Heap* heap){
    for(Sexp* r = posits; r->isCons(); r = r->cdr()){
      Sexp* posit_content = r->car();
      if(!posit_content->isCons() || 
	 !(posit_content->first()->equal(Posit::PositTagStr))){
	ITK_ERR("Illegal situation form : " << posit_content);
	throw 0;
	continue;
      }
      Posit * posit = new Posit() ;
      posit->set(posit_content, heap);
      positTable()->insert(pair<SubString, Posit*>(posit->name(), posit));
    }
  };
  
  Posit* RescueAgentBase::getPosit(const char* positname){
    PositTable::const_iterator it =
      positTable()->find(SubString(const_cast<char*>(positname)));
    if(it != positTable()->end())
      return it->second;
    else {
      ITK_ERR("error posit name" << positname);
      throw 0;
      return 0;
    }
  };

  Posit* RescueAgentBase::getPosit(SubString& positname){
    PositTable::const_iterator it = positTable()->find(positname);
    if(it != positTable()->end())
      return it->second;
    else {
      ITK_ERR("error posit name" << positname);
      throw 0;
      return 0;
    }      
  };

  Posit* RescueAgentBase::getPosit(Sexp* positname){
    PositTable::const_iterator it = positTable()->find(positname->symVal());
    if(it != positTable()->end())
      return it->second;
    else {
      ITK_ERR("error posit name" << positname);
      throw 0;
      return 0;
    }      
  };


  //--------------------------------------------------
  // RescueAgentBase::situation_add(Sexp*)
  void RescueAgentBase::situation_add(Sexp* in_positname){
    Posit* posit = getPosit(in_positname->symVal());
    if(situation().find(posit) == situation().end())
      situation().add(posit);
    else {
      ITK_ERR("already exsit this posit in situation : " << posit);
      throw 0;
    }
  };
  void RescueAgentBase::situation_remove(Sexp* out_positname){
    Posit* posit = getPosit(out_positname->symVal());
    situation().remove(posit);
  };
  void RescueAgentBase::situation_swap(Sexp* in_positname,
				       Sexp* out_positname){
    Posit* in = getPosit(in_positname->symVal());
    Posit* out = getPosit(out_positname->symVal());
    situation().remove(out);
    if(situation().find(in) == situation().end())
      situation().add(in);
    else {
      ITK_ERR("already exsit this posit in situation : " << in);
      throw 0;
    }
  };
  void RescueAgentBase::situation_clear() {
    situation().clear();
  };

};
