/*
 * Header    :
 * File      : RescueObjectPool.hxx
 * Auther    : Kosuke Shinoda
 * Since     : 2001/11/19
 * Time-stamp: <2001-12-18 13:27:29 kshinoda>
 * Comment   :
 * End       :
 */
/*
 * Copyright (C) 2001 SHINODA, Kosuke. Jaist,Japan & CARC, AIST, JAPAN
 * Copyright (C) 2001 NODA, Itsuki, CARC, AIST, JAPAN
 */

#ifndef __RescueObjectPool_hxx___
#define __RescueObjectPool_hxx___

#include <vector>
#include "../librescue/librescue.hxx"

namespace Rescue {

  class Input;
  class Object;

  /** management of set of objects **/
  class RescueObjectPool : public ObjectPool {
  private:
    
  protected:
    /** Mesh of MotionlessObject **/
    int m_meshXBase;
    int m_meshYBase;
    int m_meshXSize;
    int m_meshYSize;
    int m_meshXCount;
    int m_meshYCount;
    
    typedef vector<vector<Objects>> Mesh;
    Mesh m_mesh;
    Objects m_movingObjects;
    Objects m_firedObjects;

  public:
    virtual ~RescueObjectPool();
    RescueObjectPool();
    
    /**
     * add New Object to ObjectPool
     *  if this object already added then previous object data
     * delete from ObjectPool.
     **/
    virtual void add(Object* newObject);
    
    void addFiredObject(Object* object);
    
    void clearFiredObject();

    class Enumerator {
      friend RescueObjectPool;
      
    private:
      RescueObjectPool* m_owner;
      double m_centerX;
      double m_centerY;
      double m_squaredRadius;
      int    m_x;
      int    m_y;
      int    m_xMin;
      int    m_xMaz;
      int    m_yMin;
      int    m_yMaz;
      Objects::const_iterator m_it;
      Enumerator(RescueObjectPool* owner, S32 x, S32 y, S32 r);

      Object* nextCore();
    public:
      Object* next();
    };
    
    friend Enumerator;
    Enumerator objectsInRound(S32 x, S32, y, S32 r);

    void setUpMesh(int meshX, int mexhY);

  public:
  private:
    RescueObjectPool(const RescueObjectPool& source);
    RescueObjectPool& operator= (const RescueObjectPool& rhs);
    bool operator== (const RescueObjectPool& rhs) const;
  };

};

#endif // __RescueObjectPool_hxx___
